###########################################################################################
#
# How to add a clustering engine ?
#
##################################
#
# Simply creates a class in a new module that
# implements vizuka.clustering.clusterizer.Clusterizer
#
# Below is a simple example, with kmeans.
# If you need special parameters, pass them in vizualization.py (search "make_clusterizer")
# If these special parameters are hyperparameters, use
# self.register_parameters (see below). This is important for Vizuka to
# know which parameters to ask the user when you launch the clustering
# engine from the qt interface.
#
#
##########################################################################################


###################
#
# WORKING EXAMPLE :
#
###################
#
#class KmeansClusterizer(Clusterizer):
#
#    def __init__(self, number_of_clusters=15):
#        """
#        Uses sklearn kmeans
#        """
#        self.register_parameters(
#                parameters={'number_of_clusters':number_of_clusters})
#
#        self.method = 'kmeans'
#        self.number_of_clusters = number_of_clusters
#
#        self.engine = KMeans(
#                n_clusters=int(float(self.number_of_clusters)),
#                )
#
#    def fit(self, xs):
#        """
#        Fit the datas and find clusterization adapted to the data provided
#
#        :param xs: data to clusterize
#        """
#        self.engine.fit(xs)
#
#    def predict(self, xs):
#        """
#        Predicts cluster label
#
#        :param xs: array-like of datas
#        :return:   list of cluster possible_outputs_list
#        """
#        return self.engine.predict(xs)



import numpy as np
import logging

try:
    from MulticoreTSNE import MulticoreTSNE as tsne_algorithm
    using_multicore_tsne = True
    multicore_parameters = {'n_jobs':3}
    logging.info('dimension_reduction = Using fast tSNE')

except:
    from sklearn.manifold import TSNE as tsne_algorithm
    using_multicore_tsne = False
    multicore_parameters = {}
    logging.info('dimension_reduction = ALERT: Using slow tSNE, potential segfault if dataset too big (see requirements/requirements.apt')

from vizuka.dimension_reduction import (
        projector,
        )


class tSNE(projector.Projector):

    def __init__(self, perplexity, learning_rate, n_iter):
        
        self.method         = 'tsne'

        self.register_parameters(
                parameters = {
                    'perplexity'   : perplexity,
                    'learning_rate': learning_rate,
                    'n_iter'       : n_iter,
                    }
                )

        self.engine = tsne_algorithm(
            perplexity    = self.parameters['perplexity'],
            learning_rate = self.parameters['learning_rate'],
            n_iter        = self.parameters['n_iter'],
            **multicore_parameters,
            )
        self.projections = []

    def project(self, x):
        self.projections = self.engine.fit_transform(x)
        return self.projections



import numpy as np
from matplotlib import pyplot as plt
from sklearn.cluster import DBSCAN
from scipy.spatial import cKDTree
import logging

from vizuka.clustering.clusterizer import Clusterizer


class DBSCANClusterizer(Clusterizer):

    def __init__(self, epsilon=1.5, min_samples=30):
        """
        Inits a DBSCAN clustering engine from sklearn
        """
        self.epsilon     = float(epsilon)
        self.min_samples = int(float(min_samples))

        self.register_parameters(
                parameters={
                    'epsilon'    : self.epsilon,
                    'min_samples': self.min_samples}
                )

        self.method='dbscan'

        self.engine = DBSCAN(
                n_jobs      = 4,
                eps         = self.epsilon,
                min_samples = self.min_samples,
                )

    def fit(self, xs):
        """
        There is no dbscan.predict so...
        We are going to predict everything and
        put it on a big dict.

        This is stupid but thank sklearn for that.
        If you want to predict the class of a point
        not initially in your data (e.g the mesh_centroids)
        then the engine will first find the nearest fitted
        data, and give you its cluster labelling.

        :param xs: array-like of datas
        """
        xs_tuple = [ tuple(x) for x in xs ]
        tmp = self.engine.fit_predict(xs_tuple)
        
        self.predictions = {xs_tuple[idx]: predict for idx, predict in enumerate(tmp)}
        labels = set(tmp)

        def do(xs_tuple):
            tmp = self.engine.fit_predict(xs_tuple)
            self.predictions = {xs_tuple[idx]: predict for idx, predict in enumerate(tmp)}
            labels = set(tmp)

            f = plt.figure()
            s = f.add_subplot(111)
            s.set_title(str(len(labels))+" class")

            for i in labels:
                to_display = np.array([x for idx,x in enumerate(xs_tuple) if i == tmp[idx]])
                s.scatter(to_display[:,0], to_display[:,1])

            plt.show()
        
        # do(xs_tuple)

        self.kdtree = cKDTree(xs)
        self.xs = xs
        logging.info("DBSCAN found {} labels".format(len(labels)))

        # There is a problm here : all isolated points are classified -1
        # in DBSCAN, which is a problem for our interactive cluster selection
        # as selecting a title (labelled as the label of nearest point to its
        # "centroid") may lead to select all tiles labelled as -1 : this would
        # be very ugly

        class_min = min(labels)
        for key, class_ in self.predictions.items():
            if class_ <= -1:
                class_min-=1
                self.predictions[key] = class_min
        labels = set(self.predictions.values())
        
        logging.info("DBSCAN found {} labels after correction".format(len(labels)))

    def predict(self, xs):
        """
        Predicts cluster label
        :param xs: array-like of datas to classify
        ..seealso:: self.fit
        """
        current_predicts = []
        for x in xs:
            x_tuple = tuple(x)
            if x_tuple in self.predictions:
                current_predicts.append(self.predictions[x_tuple])
            else:
                current_predicts.append(
                    self.predictions[tuple(self.xs[self.kdtree.query(x)[1]])]
                )
        return current_predicts




import argparse
import logging
from pyfiglet import Figlet

from vizuka import dimension_reduction
from vizuka import data_loader
from vizuka import config

from vizuka.config import (
    VERSION,
    INPUT_FILE_BASE_NAME,
    DATA_PATH,
    BASE_PATH,
    PROJECTION_DEFAULT_PARAMS,
)

logger = logging.getLogger()
logger.setLevel(logging.WARN)

def do_reduce(algorithm_name, parameters, version, data_path, reduced_path):
    """
    Project the data in :param data_path: (version :param version:)
    with algorithm :param algorithm_name: using parameters in :param parameters
    and save it in :param reduced_path:
    """
    algo_builder = dimension_reduction.make_projector(algorithm_name)
    algo = algo_builder(**parameters)

    (x, _ ,_ ,_, loaded, preprocessed_filename ) = data_loader.load_preprocessed(
            file_base_name   = INPUT_FILE_BASE_NAME,
            path             = data_path,
            version          = version,
            )
    if not loaded:
        logging.warn("\nNo data found\nCorresponding file not found: {}\nPlease check --show-required-files".format(preprocessed_filename))
        return
    
    logging.warn("Projecting the preprocessed data in 2D.. this may take a while!")
    algo.project(x)                                   # do the dimension projection
    algo.save_projection(version=version, path=reduced_path)     # save the result
    logging.warn("Projecting done")


def main():
    """
    Loads parameters and run do_reduce
    """

    print(Figlet().renderText('Vizuka'))

    builtin_projectors, extra_projectors = dimension_reduction.list_projectors()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-a', '--algorithm',
        help=(
            'algorithm name to reduce dimensions, available are :\n'
             + 'builtin: {}\n'.format(list(builtin_projectors.keys()))
             + 'extra plugins: {}'.format(list(extra_projectors.keys()))
             )
         )
    parser.add_argument(
        '-p', '--parameters', action='append',
        help='specify parameters, e.g: "-p perplexity:50 -p learning_rate:0.1" '
             'It will load default values in config.py if not specified'
        )
    parser.add_argument(
        '-v', '--version',
        help='specify a version of the files to load/generate (see vizuka --show_required_files), currently: '+VERSION)
    parser.add_argument(
        '--path',
        help='change the location of your data/ folder containing reduced/ (the projections)'
             ' and set/ (the raw and preprocessed data).'
             ' Default to {}'.format(BASE_PATH)
        )
    parser.add_argument(
            '--verbose', action="store_true",
            help="verbose mode")

    parser.set_defaults(
            algorithm = 'manual',
            parameters  = {},
            path        = BASE_PATH,
            version     = VERSION,
            )

    args = parser.parse_args()

    (data_path, reduced_path, _, _, _, _) = config.path_builder(args.path)

    algorithm_name = args.algorithm
    parameters     = args.parameters
    version        = args.version
    verbose        = args.verbose

    print("VERSION: Loading dataset labeled {} (cf --version)".format(version))
    
    if verbose:
        logger.setLevel(logging.DEBUG)
    
    if algorithm_name == 'manual':
        choice_list, choice_dict = "", {}
        for i,method in enumerate([*builtin_projectors, *extra_projectors]):
            choice_list+="\t\t [{}]: {}\n".format(i, method)
            choice_dict[i]=method
        choice = input("No algorithm specified (-a or --algorithm)\n\tChoices available are:\n"+choice_list+"\t[?] > ")
        try:
            choice_int=int(choice)
        except:
            logger.warn("Please enter a valid integer !")
            return
        algorithm_name = choice_dict[choice_int]
              

    for raw_param in parameters:
        if ':' not in raw_param:
            raise TypeError('parameter -p not used correctly! see --help')
        param_name, param_value = raw_param.split(':')
        parameters[param_name:param_value]

    default_params = PROJECTION_DEFAULT_PARAMS.get(algorithm_name, {})
    for default_param_name, default_param_value in default_params.items():
        if default_param_name not in parameters:
            parameters[default_param_name] = default_param_value
            logger.info('parameter {} not specified, using default value: {}'.format(
                                default_param_name, default_param_value)
                            )

    do_reduce(algorithm_name, parameters, version, data_path, reduced_path)
    
    cmd_to_launch = "vizuka"
    cmd_to_launch+= (' --version {}'.format(version) if version!=VERSION else '')
    cmd_to_launch+= ('--path {}'.format(args.path) if args.path != BASE_PATH else '')
    print('"Projection done, you can now launch "vizuka"')

if __name__=='__main__':
    main()



import os
import pkgutil
import inspect
import importlib
import pickle
import logging

from vizuka.clustering import (
        kMeans,
        dbscan,
        grid,
        clusterizer,
        )

from vizuka.plugins import clustering as clustering_plugins

def get_required_arguments(method_name):
    if method_name=='grid':
        return []
    else:
        builtin, extra = list_clusterizer()
        available_clusterizers = {**builtin, **extra}
        return inspect.getargspec(available_clusterizers[method_name]).args[1:]

def list_clusterizer():
    built_in_clusterizers = {
            'kmeans':kMeans.KmeansClusterizer,
            'dbscan':dbscan.DBSCANClusterizer,
            'grid':grid.DummyClusterizer,
            }

    extra_cluterizers = {}
    
    for (module_loader, name, ispkg) in pkgutil.iter_modules(clustering_plugins.__path__):
        plugin = importlib.import_module('.plugins.clustering.'+name, 'vizuka')
        members = inspect.getmembers(plugin)
        for _, class_ in members:
            if class_ in clusterizer.Clusterizer.__subclasses__():
                extra_cluterizers[name] = class_

    return built_in_clusterizers, extra_cluterizers
    

def make_clusterizer(method='grid', **kwargs):
    """
    Clusterize the data with specified algorithm
    Naively assume you pass the right parameters for the right algo

    :param data:       array with shape (n,2) of inputs to clusterize
    :param method:     algo to use, supported: kmeans, dbscan,grid 
    :param n_clusters: nb of clusters to find (if applicable)

    :return: a clusterizer object (instance of child of Clusterizer())
    """
    
    built_in_clusterizers, extra_clusterizers = list_clusterizer()
    available_clusterizers = {**built_in_clusterizers, **extra_clusterizers}
    
    clusterizer_builder = available_clusterizers.get(method, None)
    
    required_parameters = inspect.getargspec(clusterizer_builder).args
    given_parameters    = {name:value for name,value in kwargs.items() if name in required_parameters}
    clusterizer = clusterizer_builder(**given_parameters)
    
    return clusterizer



import os
import logging

import numpy as np

from vizuka.config import (
        VERSION,
        REDUCED_DATA_PATH,
        NAME_VALUE_SEPARATOR,
        PARAMETERS_SEPARATOR,
        )

class Projector():

        def __init__(self):
            """
            Instanciate a Projector engine, you should give it
            here all the parameters
            """
            self.method='null'
            self.projections = np.array([])
            self.parameters = {} # In your implementation, do add you parameters here !
                                 # If you don't, two projections with different
                                 # parameters will be saved in the same file, the new
                                 # one overwritting the old one ..see: self.save
        
        def project(self, x):
            """
            Projects the x vector into a 2D space according to the
            parameters specified in __init__
            """
            self.projections = np.array([[0,0] for _ in range(len(x))])
            return self.projections

        # GENERIC PROJECTOR FUNCTIONS
        #############################

        def register_parameters(self, parameters):
            """
            This function looks silly but it enforces the fact that parameters should
            alml remain in the same dict for some other functions to be able to fetch
            them by name:value
            """
            self.parameters = parameters

        def get_param_from_name(name):
            """
            Returns the method and parameters of the projector
            from its name
            """
            params  = {}
            blocks  = name.split(PARAMETERS_SEPARATOR)
            method, version, *blocks = blocks
            
            for block in blocks:
                if block:
                    name, value  = block.split(NAME_VALUE_SEPARATOR)
                    params[name] = value
            
            return method, version, params

        def get_savename(self, version=VERSION, path=REDUCED_DATA_PATH):
            name = ''.join([self.method, PARAMETERS_SEPARATOR, version])
            ordered_name = np.sort(list(self.parameters.keys()))
            name += ''.join(PARAMETERS_SEPARATOR + str(param_name) + NAME_VALUE_SEPARATOR + str(self.parameters[param_name]) for param_name in ordered_name)
            return ''.join([path, name, '.npz'])
     
        def load_projection(self, version=VERSION, path=REDUCED_DATA_PATH):
            full_path = self.get_savename(version, path)
            if os.path.exists(full_path):
                return np.load(full_path)['x_2D']
            else:
                logging.info("Nothing found in {}!".format(full_path))
                return np.array([])

        def save_projection(self, version=VERSION, path=REDUCED_DATA_PATH):
            """
            Save the projection in standard location with appropriate name,
            which depends on the parameters of the algorithm
            """
            full_path = self.get_savename(version, path)
            np.savez(full_path, x_2D=self.projections)



"""
Module built around the class Vizualization
It is the main module that draws a nice IHM interface to explore your data

See class Vizualization
"""
import sys
import os
import logging
from collections import Counter
import pickle
import csv

from scipy import stats
import numpy as np
import matplotlib
matplotlib.use('Qt5Agg')  # noqa
from matplotlib.gridspec import GridSpec
from matplotlib import pyplot as plt

from vizuka.helpers import viz_helper
from vizuka.graphics import drawing
from vizuka import clustering
from vizuka import frontier
from vizuka import data_loader
from vizuka import heatmap
from vizuka.graphics.qt_handler import Cluster_viewer
from vizuka.graphics.qt_handler import Viz_handler
from vizuka.config import (
        path_builder,
        BASE_PATH,
        )


class Vizualization:

    """
    This class contains all the tools for the vizualiation

    It is created with only the vector of predictions (containing possible_outputs_list) with its decoder function
    Resolution is the size of the grid of the viz, used for heatmap and clustering
    Mouse control and keyboard shortcuts are used (later: QtButtons) ..seealso:: self.controls
    """
    
    def __init__(
            self,
            raw_inputs,
            raw_inputs_columns,
            projected_input,
            predicted_outputs,
            correct_outputs,
            resolution=100,
            special_class='0',
            nb_of_clusters=120,
            features_name_to_filter=[],
            features_name_to_display={},
            heatmaps_requested = ['entropy', 'accuracy'],
            color_mode = 'by_prediction',
            class_decoder=(lambda x: x), class_encoder=(lambda x: x),
            output_path='output.csv',
            base_path=BASE_PATH,
            version='',
            ):
        """
        Central function, draw heatmap + scatter plot + zoom + annotations on tSNE data

        :param predicted_outputs: vector of predicted possible_outputs_list
        :param correct_outputs: vector of true possible_outputs_list
        :param projected_input: vector of t-SNE projected data (i.e project_input[index] = (x, y)
        :param class_dn_jobs=4, ecoder: function to decode possible_outputs_list machinetohuman
        :param class_encoder: dict to encode possible_outputs_list humantomachine
        """

        logging.info("Vizualization=generating")
        self.last_time = {}

        self.manual_cluster_color = 'cyan'
        
        (
                self.data_path,
                self.reduced_data_path,
                self.model_path,
                self.graph_path,
                self.cache_path, 
                self.saved_clusters_path,
                
                ) = path_builder(base_path)

        self.predictors = [name for name in os.listdir(self.model_path) if version in name]
        self.output_path = output_path
        self.saved_clusters = [f for f in os.listdir(self.saved_clusters_path) if version in f]
        self.version = version
        
        def str_with_default_value(value):
            if not value:
                return str(special_class)
            return str(value)

        self.correct_outputs = [str_with_default_value(correct_output) for correct_output in correct_outputs]
        self.correct_outputs_original = self.correct_outputs
        self.prediction_outputs = [str_with_default_value(predicted_output) for predicted_output in predicted_outputs]
        self.prediction_outputs_original = self.prediction_outputs
        
        self.projected_input = projected_input
        self.projected_input_original = self.projected_input
        self.x_raw = raw_inputs
        self.x_raw_original = self.x_raw
        self.x_raw_columns = raw_inputs_columns
        self.nb_of_clusters = nb_of_clusters
        self.class_decoder = class_decoder
        self.special_class = str(special_class)
        self.heatmaps_requested = heatmaps_requested
        
        self.features_name_to_filter = features_name_to_filter
        self.correct_class_to_display = {}
        self.predicted_class_to_display = {}
        self.feature_to_display_by_col = {}
        self.features_to_display = features_name_to_display
        self.color_mode = color_mode
        self.cluster_view_selected_indexes = []
        self.left_clicks = set()
        self.selected_clusters = set()

        #self.possible_outputs_list = list({self.class_decoder(y_encoded) for y_encoded in self.correct_outputs})
        self.possible_outputs_set = set(self.correct_outputs).union(set(self.prediction_outputs))
        self.possible_outputs_set.discard(None)
        self.filters = {
                'PREDICTIONS':self.possible_outputs_set,
                'GROUND_TRUTH':self.possible_outputs_set,
                **{k:set() for k in features_name_to_filter},
                }
        self.possible_outputs_list = list(self.possible_outputs_set)
        self.possible_outputs_list.sort()
        # logging.info("correct outputs : %", self.correct_outputs)
        self.projection_points_list_by_correct_output = {y: [] for y in self.correct_outputs}
        self.index_by_true_output = {class_:[] for class_ in self.possible_outputs_list}
        self.nb_of_individual_by_true_output = {}

        for index, projected_input in enumerate(self.projected_input):
            self.projection_points_list_by_correct_output[self.correct_outputs[index]].append(projected_input)
            self.index_by_true_output[self.correct_outputs[index]].append(index)

        # convert dict values to np.array
        for possible_output in self.projection_points_list_by_correct_output:
            self.projection_points_list_by_correct_output[possible_output] = np.array(
                self.projection_points_list_by_correct_output[possible_output])
            self.nb_of_individual_by_true_output[possible_output] = len(
                self.projection_points_list_by_correct_output[possible_output])
        self.projection_points_list_by_correct_output_original = self.projection_points_list_by_correct_output

        self.resolution = resolution # counts of tiles per row/column

        self.shift_held = False
        self.ctrl_held = False

        self.local_effectif = {}
        self.local_accuracy = {}
        self.local_confusion_by_class = {
                class_:
                {
                    class__:0 for class__ in self.possible_outputs_list} for class_ in self.possible_outputs_list
                }
        self.local_bad_count_by_class = {class_:0 for class_ in self.possible_outputs_list}
        self.local_classes = set()
        self.local_sum = 0
        self.cursor_ids = [0]

        # Get the real possible_outputs_list found in true y

        self.amplitude = viz_helper.find_amplitude(self.projected_input_original)

        mesh = np.meshgrid(
                np.arange(
                    -self.amplitude,
                    self.amplitude,
                    self.amplitude/(self.resolution/2)
                    ),
                np.arange(
                    -self.amplitude,
                    self.amplitude,
                    self.amplitude/(self.resolution/2)
                    ),
                )
        
        self.size_centroid  = 2 * self.amplitude / self.resolution
        self.mesh_centroids = np.c_[mesh[0].ravel(), mesh[1].ravel()]

        self.calculate_prediction_projection_arrays()

        logging.info('clustering engine=fitting')
        self.clusterizer = clustering.make_clusterizer(
                xs=self.projected_input_original,
                nb_of_clusters=self.nb_of_clusters,
                mesh=self.mesh_centroids,
                method='grid',
                )
        logging.info('clustering engine=ready')
        self.normalize_frontier = True
        
    def calculate_prediction_projection_arrays(self):
        """
        Update all the sets which use true/false flags from predictions
        When to use ? -> if you change (or first load) your predictor
        """

        (
            self.index_bad_predicted,
            self.index_good_predicted,
            self.index_not_predicted,
        ) = viz_helper.separate_prediction(
                self.prediction_outputs,
                self.correct_outputs,
                self.special_class,
                )

        (
            self.well_predicted_projected_points_array,
            self.mispredicted_projected_points_array,
            self.not_predicted_projected_points_array,

            ) = viz_helper.get_projections_from_index(
                    self.index_good_predicted,
                    self.index_bad_predicted,
                    self.index_not_predicted,
                    self.projected_input,
                    )

        self.accuracy_by_class = viz_helper.get_accuracy_by_class(
                self.index_by_true_output,
                self.correct_outputs,
                self.prediction_outputs,
                self.possible_outputs_list)


    def reload_predict(self, filename):
        """
        Call this function if the predictor set has changed
        filename: the name of the predictions file to load, should
        be located in the self.model_path folder
        """
        old_left_clicks = self.left_clicks

        self.prediction_outputs = data_loader.load_predict_byname(filename, path=self.model_path)
        self.prediction_outputs_original = self.prediction_outputs

        self.correct_outputs = self.correct_outputs_original
        self.projected_input = self.projected_input_original

        self.calculate_prediction_projection_arrays()

        self.print_global_summary(self.global_summary_axe)

        self.init_clusters()
        self.update_all_heatmaps()

        self.reset_summary()
        self.reset_viz()
        self.conciliate_filters(self.filters)
        for click in old_left_clicks:
            self.do_left_click(click)
        
        self.refresh_graph()


    def filter_by_feature(self, feature_col, selected_features):
        """
        Updates the list of index to display, filter with :

        feature_col: the column of the feature in "originals"
        selected_feature_list: list of checked/unchecked feature to display
        """
        featured_data_to_display = {
                widget.class_ for widget in selected_features[feature_col] if widget.isChecked()
                }
        self.filters[feature_col] = featured_data_to_display
        self.conciliate_filters(self.filters)

    def filter_by_feature_old(self, feature_col, selected_feature_list):
        """
        Updates the list of index to display, filter with :

        feature_col: the column of the feature in "originals"
        selected_feature_list: list of checked/unchecked feature to display
        """

        featured_data_to_display = [
                item for item, selected in selected_feature_list.items() if selected
                ]
        self.feature_to_display_by_col[feature_col] = featured_data_to_display
        self.filters[feature_col] = featured_data_to_display

        # self.display_by_filter()
        self.conciliate_filters(self.filters)

    def filter_by_correct_class(self, selected_outputs_class_list):
        """
        Filter by class, on the TRUE class of each point
        """
        self.correct_class_to_display = {
                widget.class_ for widget in selected_outputs_class_list if widget.isChecked()
                }
        self.filters["GROUND_TRUTH"] = self.correct_class_to_display
        # self.display_by_filter()
        self.conciliate_filters(self.filters)

    def filter_by_predicted_class(self, selected_outputs_class_list):
        """
        Filter by class, on the PREDICTED class of each point
        """
        self.predicted_class_to_display = {
                widget.class_ for widget in selected_outputs_class_list if widget.isChecked()
                }
        self.filters["PREDICTIONS"] = self.predicted_class_to_display
        # self.display_by_filter()
        self.conciliate_filters(self.filters)

    def load_only_some_indexes(self, indexes):
        self.prediction_outputs = [self.prediction_outputs_original[i]  for i in indexes]
        self.projected_input    = [self.projected_input_original[i]     for i in indexes]
        self.correct_outputs    = [self.correct_outputs_original[i]     for i in indexes]
        self.x_raw              = [self.x_raw_original[i]               for i in indexes]

        self.projection_points_list_by_correct_output = {y: [] for y in self.correct_outputs}
        self.index_by_true_output = {class_:[] for class_ in self.possible_outputs_list}
        
        for index, projected_input in enumerate(self.projected_input):
            self.projection_points_list_by_correct_output[self.correct_outputs[index]].append(projected_input)
            self.index_by_true_output[self.correct_outputs[index]].append(index)

        # convert dict values to np.array
        for possible_output in self.projection_points_list_by_correct_output:
            self.projection_points_list_by_correct_output[possible_output] = np.array(
                self.projection_points_list_by_correct_output[possible_output])
            self.nb_of_individual_by_true_output[possible_output] = len(
                self.projection_points_list_by_correct_output[possible_output])

        self.calculate_prediction_projection_arrays()
        self.init_clusters()
        
        self.print_global_summary(self.global_summary_axe)
        self.update_all_heatmaps()
        self.update_summary()
        self.print_summary(self.summary_axe)
        self.update_cluster_view()
    
    def filters_active(self):
        """
        True if some filters are active
        """

        additional_filters_active = False
        other_filters = {k:v for k,v in self.filters.items() if k!="PREDICTIONS" and k!="GROUND_TRUTH"}
        for col in other_filters: # other filters are column number identifying features
            additional_filters_active = additional_filters_active or other_filters[col]

        return self.filters["GROUND_TRUTH"] or self.filters["PREDICTIONS"] or additional_filters_active
        

    def conciliate_filters(self, filters):
        saved_zoom = (self.ax.get_xbound(), self.ax.get_ybound())
        to_display = set(range(len(self.projected_input_original)))

        filtered = filters["GROUND_TRUTH"]
        to_display = to_display.intersection(set([
                idx for idx, class_ in enumerate(self.correct_outputs_original) if class_ in filtered]))
        
        filtered = filters["PREDICTIONS"]
        to_display = to_display.intersection(set([
                idx for idx, class_ in enumerate(self.prediction_outputs_original) if class_ in filtered]))
        other_filters = {k:v for k,v in filters.items() if k!="PREDICTIONS" and k!="GROUND_TRUTH"}
        for col in other_filters: # other filters are column number identified features
            filtered_by_feature = {idx for idx, x in enumerate(self.x_raw_original) if x[col] in other_filters[col]}
            to_display = to_display.intersection(filtered_by_feature)

        viz_helper.remove_pathCollection(self.ax)
        
        self.load_only_some_indexes(to_display)

        drawing.draw_scatterplot(
                self,
                self.ax,
                color_mode = self.color_mode,
                )
        self.ax.set_xbound(saved_zoom[0])
        self.ax.set_ybound(saved_zoom[1])

        self.refresh_graph()

    #######################################
    # Similarity functions to draw clusters

    def onclick(self, event):
        """
        Mouse event handler

        Actions on mouse button pressed
            1 : select a tile (and a class)
            2 : find similar tiles
            3 : reset vizualization (graph+summary)
            
        """
        x = event.xdata
        y = event.ydata
        button = event.button
        left_click, right_click = ((button == 1), (button == 3))

        self.summary_axe.clear()
        self.summary_axe.axis('off')

        if left_click:
            self.do_left_click((x,y))

        elif right_click:
            self.do_right_click((x,y))

        self.refresh_graph()
    
    def do_left_click(self, xy):
        x, y = xy
        if (x is None) or (y is None):  # clicks out of the screen
            return
        
        # Register this click for replay
        self.left_clicks.add((x,y))
        
        # find associated cluster and gather data
        clicked_cluster = self.clusterizer.predict([(x,y)])[0]
        self.selected_clusters.add(clicked_cluster)
        self.delimit_cluster(clicked_cluster, color=self.manual_cluster_color)

        self.update_summary()
        self.print_summary(self.summary_axe)
        
        # shows additional info if requested (argument -s)
        self.update_cluster_view()

            # self.cluster_view.update_cluster_view(
            #        clicked_cluster,
            #        self.index_by_cluster_label,
            #        indexs_good = self.index_good_predicted,
            #        indexes_bad = self.index_bad_predicted,
            #        )

    def do_right_click(self, xy):
            logging.info("Cleaning state, unselecting clusters, redrawing")
            # reboot vizualization
            self.left_clicks = set()
            self.selected_clusters = set()
            self.reset_summary()
            self.clear_cluster_view()
            self.reset_viz()

    def reset_viz(self):
        """
        Reset (graphically) the vizualization
        ..note:: does not touch the summary array, for this use self.reset_summary()
        """
        logging.info("scatterplot: removing specific objects")
        
        self.left_clicks = set()
        self.selected_clusters = set()
        for ax in self.axes_needing_borders:
            for i in ax.get_children():
                if isinstance(i, matplotlib.collections.PathCollection):
                    i.remove()
                elif isinstance(i, matplotlib.lines.Line2D):
                    if i.get_color() == self.manual_cluster_color:
                        i.remove()
        
        logging.info("scatterplot: drawing observations")
        if self.cluster_view:
            self.clear_cluster_view()

        drawing.draw_scatterplot(
                self,
                self.ax,
                color_mode = self.color_mode,
                )
        logging.info("scatterplot: ready")


    def reset_summary(self):
        """
        Reset the local summary
        """
        self.local_effectif = {}
        self.local_accuracy = {}
        self.local_confusion_by_class = {
                output_class: {
                    other_output_class:0 for other_output_class in self.possible_outputs_list
                    }
                for output_class in self.possible_outputs_list
                }
        self.local_bad_count_by_class = {
                output_class:0 for output_class in self.possible_outputs_list
                }
        self.local_classes = set()
        self.local_sum = 0
        self.selected_clusters = set()

        self.summary_axe.clear()
        self.summary_axe.axis('off')

    def init_clusters(self):
        """
        Clusterize the data and retrieve some countings
        """
        (
            self.index_by_cluster_label,
            self.all_cluster_labels,
            self.nb_of_points_by_class_by_cluster,
            self.cluster_by_idx,

            ) = viz_helper.cluster_label_mesh(

                self.clusterizer,
                self.projected_input,
                self.projection_points_list_by_correct_output_original,
                self.correct_outputs,
                )


        (
            self.nb_good_point_by_cluster,
            self.nb_bad_point_by_cluster,
            self.nb_good_point_by_class_by_cluster,
            self.nb_bad_point_by_class_by_cluster,
            self.nb_null_point_by_cluster,

            ) = viz_helper.get_count_by_cluster(

                self.all_cluster_labels,
                self.index_by_cluster_label,
                self.correct_outputs,
                self.index_good_predicted,
                self.index_bad_predicted,
                self.index_not_predicted,
                )


    def calculate_centroid_coordinates(self, x, y):
        return x + self.resolution * y

    def draw_the_line(self, x_list, y_list, color='b'):
        for axe in self.axes_needing_borders:
            axe.add_artist(matplotlib.lines.Line2D(xdata=x_list, ydata=y_list, color=color))

    def line(self, float_point):
        return (float_point - self.size_centroid / 2, float_point + self.size_centroid / 2)

    def lower_bound(self, float_point, plus_one=None):
        if plus_one:
            return (float_point + self.size_centroid / 2,)
        return (float_point - self.size_centroid / 2,)


    def delimit_cluster(self, cluster, color='b', **kwargs):
        """
        Delimits one cluster by drawing lines around it
        """
        centroids_cluster_by_index = self.clusterizer.predict(self.mesh_centroids)

        cluster_y_list_by_x = [[] for x in range(self.resolution)]
        cluster_x_list_by_y = [[] for y in range(self.resolution)]

        iter_all_coordinates = ((x, y) for x in range(0, self.resolution) for y in range(0, self.resolution))

        for idx, (x, y) in enumerate(iter_all_coordinates):
            if centroids_cluster_by_index[idx] == cluster:
                cluster_y_list_by_x[x].append(y)
                cluster_x_list_by_y[y].append(x)
                
        calculate_coordinates = self.calculate_centroid_coordinates

        def draw_all_lines(self, list_points, swapped_coordinates=False):
            for a in range(0, self.resolution):
                a_line_b_list = list_points[a]
                if not a_line_b_list:
                    continue
                min_b = min(a_line_b_list)
                max_b = max(a_line_b_list)
                no_hole = (max_b - min_b + 1) == len(a_line_b_list)
                if no_hole:
                    if swapped_coordinates:  # swaped if a is y and b is x, not swapt if a is  and b is y
                        b_float_position, a_float_position= self.mesh_centroids[calculate_coordinates(min_b, a)]
                        self.draw_the_line(
                                self.lower_bound(b_float_position),
                                self.line(a_float_position),
                                color=color)
                        b_float_position, a_float_position = self.mesh_centroids[calculate_coordinates(max_b, a)]
                        self.draw_the_line(
                            self.lower_bound(b_float_position, plus_one=True),
                            self.line(a_float_position),
                            color=color)
                    else:
                        a_float_position, b_float_position= self.mesh_centroids[calculate_coordinates(a, min_b)]
                        self.draw_the_line(
                                self.line(a_float_position),
                                self.lower_bound(b_float_position),
                                color=color)
                        a_float_position, b_float_position = self.mesh_centroids[calculate_coordinates(a, max_b)]
                        self.draw_the_line(
                            self.line(a_float_position),
                            self.lower_bound(b_float_position, plus_one=True),
                            color=color)
                else:  # case not convex, which is not often so it's gonna be dirty
                    for b in a_line_b_list:
                        if swapped_coordinates:
                            if (b - 1) not in a_line_b_list:
                                b_float_position, a_float_position = self.mesh_centroids[calculate_coordinates(b, a)]
                                self.draw_the_line(
                                        self.lower_bound(b_float_position),
                                        self.line(a_float_position),
                                        color=color)
                            if (b + 1) not in a_line_b_list:
                                b_float_position, a_float_position = self.mesh_centroids[calculate_coordinates(b, a)]
                                self.draw_the_line(
                                    self.lower_bound(b_float_position, plus_one=True),
                                    self.line(a_float_position),
                                    color=color)
                        else:
                            if (b - 1) not in a_line_b_list:
                                a_float_position, b_float_position = self.mesh_centroids[calculate_coordinates(a, b)]
                                self.draw_the_line(
                                        self.line(a_float_position),
                                        self.lower_bound(b_float_position),
                                        color=color)
                            if (b + 1) not in a_line_b_list:
                                a_float_position, b_float_position = self.mesh_centroids[calculate_coordinates(a, b)]
                                self.draw_the_line(
                                    self.line(a_float_position),
                                    self.lower_bound(b_float_position, plus_one=True),
                                    color=color)
                            
        draw_all_lines(self, cluster_y_list_by_x, swapped_coordinates=True)
        draw_all_lines(self, cluster_x_list_by_y, swapped_coordinates=False)

    

    def get_coordinates_from_index(self, index):
        return (self.resolution - index // self.resolution -1, index % self.resolution)
        # check it's the same
        # return (self.resolution - int(((index - index % self.resolution) / self.resolution)) - 1,
        #         index % self.resolution)

    def request_color_mode(self, mode):
        self.color_mode = mode
        try:
            drawing.draw_scatterplot(
                    self,
                    self.ax,
                    color_mode=self.color_mode
                    )
            self.refresh_graph()
        except:
            pass

    def request_new_frontiers(self, method):
        """
        Init new frontiers based on a similarity mesure between clusters
        """
        method = method.lower()
        logging.info('frontiers : requiring new delimitations '+method)

        for axe in self.axes_needing_borders:
            for child in axe.get_children():
                if isinstance(child, matplotlib.collections.LineCollection):
                    child.remove()
                    logging.info("removing a line collection")

        similarity = frontier.make_frontier(method)
        self.similarity_measure = similarity.compare

        if method == 'bhattacharyya':
            self.normalize_frontier=True
        elif method =='all':
            self.normalize_frontier=False
        elif method == 'none':
            self.normalize_frontier=False
            self.refresh_graph()
            return
        
        drawing.apply_borders(
                self,
                self.normalize_frontier,
                self.similarity_measure,
                self.axes_needing_borders
                )
        self.refresh_graph()
        logging.info('frontiers : applied '+method)

    def request_clusterizer(self, method, params):
        """
        Creates the clustering engine depending on the method you require
        Actually different engine use different initialisations

        :param params: the parameters as a dict, of the clustering
        :param method: the name of the clustering engine
        """
        logging.info("cluster: requesting a new " + method + " engine")
        self.clusterizer = clustering.make_clusterizer(
                method = method,
                mesh = self.mesh_centroids,
                **params,
                )

        if (not self.filters_active()) and self.clusterizer.loadable(self.version, self.cache_path):
            logging.info("cluster: found cached clustering, loading..")
            self.clusterizer = self.clusterizer.load_cluster(self.version, self.cache_path)
            logging.info("cluster: ready")
        else:
            logging.info("cluster: fitting the clusters")
            self.clusterizer.fit(self.projected_input_original)
            if not self.filters_active():
                logging.info("cluster: saving the clusters")
                self.clusterizer.save_cluster(self.version, self.cache_path)
            logging.info("cluster: ready")
            
    def request_new_clustering(self, method, params):
        """
        Init and fit a new clustering engin, then update the heatmaps

        :param method: clustering engine to use ..seealso:clustering module
        :param params: a dict with the parameters
        """
        saved_zoom = (self.ax.get_xbound(), self.ax.get_ybound())

        method = method.lower()
        self.request_clusterizer(method, params)

        self.init_clusters()
        self.update_all_heatmaps()

        drawing.apply_borders(
                self,
                self.normalize_frontier,
                self.similarity_measure,
                self.axes_needing_borders)
        logging.info('borders: done')


        self.reset_summary()
        self.reset_viz()
        self.ax.set_xbound(saved_zoom[0])
        self.ax.set_ybound(saved_zoom[1])
        self.refresh_graph()

    def update_all_heatmaps(self):
        """
        Get all heatmaps registered by request_heatmap and draw them from scratch
        """
        for (this_heatmap, axe, title) in self.heatmaps:
            axe.clear()
            axe.axis('off')

            this_heatmap.update_colors(vizualization=self)
            heatmap_color = this_heatmap.get_all_colors()

            im = axe.imshow(
                    heatmap_color,
                    interpolation='nearest',
                    vmin=0, vmax=1,
                    extent=(
                        -self.amplitude-self.size_centroid/2,
                        self.amplitude-self.size_centroid/2,
                        -self.amplitude-self.size_centroid/2,
                        self.amplitude-self.size_centroid/2
                        ),
                    aspect='auto')

            axe.set_xlim(-self.amplitude / 2, self.amplitude / 2)
            axe.set_ylim(-self.amplitude / 2, self.amplitude / 2)
            axe.axis('off')
            axe.set_title(title)

        self.refresh_graph()

    def update_summary(self):
        """
        Updates the local variables (the variables containing info about currently
        selected clusters).

        Three objects are important inside the object Vizualization and need to be updated :
            - self.local_classes contains possible_outputs_list inside current_cluster
            - self.local_effectif contains the effetif of each label inside current_cluster
            - self.local_accuracy is the ratio of good/total predicted inside cluster, by label
            - self.local_confusion_by_class is the error made by the predictor, indexed by true output
        """
        effectif_by_class = {}
        logging.info("local_variables=updating")
        
        # Get effectif
        for clr in self.selected_clusters:
            for cls, effectif in self.nb_of_points_by_class_by_cluster.get(clr,{}).items():
                effectif_by_class[cls] = effectif_by_class.get(cls,0) + effectif
        effectif_by_class = {cls:e for cls, e in effectif_by_class.items() if e>0}

        # Set 3 first local variables
        self.local_classes  = list(effectif_by_class.keys())
        self.local_sum      = sum (effectif_by_class.values())
        self.local_effectif = effectif_by_class

        # Counts the good/bad predictions in these clusters
        logging.info("local_variables=counting good/bad")
        nb_good_by_class = {cls:sum([self.nb_good_point_by_class_by_cluster.get(clr,{}).get(cls,0) for clr in self.selected_clusters]) for cls in self.local_classes}
        nb_bad_by_class  = {cls:sum([self.nb_bad_point_by_class_by_cluster.get(clr,{}).get(cls, 0) for clr in self.selected_clusters]) for cls in self.local_classes}

        self.local_bad_count_by_class  = nb_bad_by_class
        self.local_good_count_by_class = nb_good_by_class
        
        # Gets accuracy
        logging.info("local_variables=counting accuracy")
        for cls in self.local_classes:
            self.local_accuracy[cls] = (
                nb_good_by_class.get(cls,0) /
                (
                    nb_good_by_class.get(cls,0)
                    + nb_bad_by_class.get(cls,0)
                    ))
        
        # Get confusion matrix
        logging.info("local_variables=calculating confusion")
        confusion = {}
        for clr in self.selected_clusters:
            for idx in self.index_by_cluster_label.get(clr,{}):
                if idx in self.index_bad_predicted:
                    correct            = self.correct_outputs[idx]
                    prediction         = self.prediction_outputs[idx]
                    confusion[correct] = confusion.get(correct, []) + [prediction]
        
        self.local_confusion_by_class_sorted = {cls:[] for cls in confusion}
        for cls, errors in confusion.items():
            self.local_confusion_by_class_sorted[cls] = Counter(errors).most_common(2)
        logging.info("local_variables=ready")


    def print_summary(self, axe, max_row=15):
        """
        Print a short summary with basic stats (occurrence, classification rate) on an axe

        This method gets its data mainly from self.local_accuracy and self.local_effectif,
        these objects are self-updated when needed and contain the data of the user-selected clusters
        
        :param max_row: max nb of row to add in table summary
        :param axe: the matplotlib axe in which the stats will be plotted
        """
        logging.info("local_summary=printing")
        row_labels = list(self.local_classes)

        values = [
            [
                (
                    '{0:.0f} ({1:.1f}%)'.format(
                        self.local_effectif[c],
                        self.local_effectif[c] / self.nb_of_individual_by_true_output[c] * 100)
                    ),
                (
                    '{0:.2f}% ({1:+.1f}%) - {0:.1f}'.format(
                        self.local_accuracy[c]*100,
                        (self.local_accuracy[c]-self.accuracy_by_class[c])*100,
                        stats.binom_test(
                            self.local_effectif[c]*self.local_accuracy[c],
                            self.local_effectif[c],
                            self.accuracy_by_class[c],
                            alternative='two-sided'
                            ),
                        )
                    ),
                (
                    '{0:.0f} ({1:.1f}%)'.format(
                        self.nb_of_individual_by_true_output[c],
                        self.nb_of_individual_by_true_output[c] / float(len(self.projected_input)) * 100,
                        )
                    ),
                (
                    ' '.join([
                        '{:.6}'.format(class_mistaken)+ ' ({0:.1f}%)'.format(
                            error_count/float(self.local_bad_count_by_class[c])*100
                            )
                        for class_mistaken, error_count in self.local_confusion_by_class_sorted.get(c,[]) if error_count != 0
                        ])
                    ),
            ]
            for c in row_labels
        ]

        arg_sort = np.argsort([self.local_effectif[c] for c in row_labels])

        values = [values[i] for i in arg_sort[::-1]]
        row_labels = [row_labels[i][:6] for i in arg_sort[::-1]]
        # add row "all" for recap :

        max_row    = min(max_row, min(len(values), len(row_labels)))
        values     = values[:max_row]
        row_labels = row_labels[:max_row]
        
        values.append([self.local_sum, '-', '{} (100%)'.format(len(self.projected_input)), ' '])
        row_labels.append('all')

        self.rows = row_labels

        cols = [
                '#class (in cluster)\n'
                '(#class (in cluster) /#class)',
                'accuracy (in cluster)\n'
                '(delta with accuracy (general)) - p-value',
                '#class\n'
                '(#class /#all_class)',
                'common mistakes'
                ]
        
        axe.clear()
        summary = axe.table(
            cellText=values,
            rowLabels=row_labels,
            colLabels=cols,
            loc='center',
        )

        cells = summary.get_celld()
        for i in range(0,4):
            cells[(0,i)].set_height(cells[(1,0)].get_height()*2)

        
        summary.auto_set_font_size(False)
        summary.set_fontsize(8)
        axe.axis('off')

        logging.info("local_summary=ready")
        
    def print_global_summary(self, ax, max_row=9):
        """
        Prints a global summary with the most frequent class and
        their classification accuracy
        """
        logging.info("global_summary=calculating")
        
        ax.clear()
        ax.axis('off')
        cols = ['accuracy', 'effectif']
        most_common_classes = Counter(
                {
                    c:len(self.index_by_true_output.get(c,[])) for c in self.possible_outputs_list
                    }
                ).most_common(max_row)

        row_labels = np.array(most_common_classes)[:,0]
        values = []
        for c in row_labels:
            accuracy = self.accuracy_by_class.get(c,0)*100
            effectif = self.nb_of_individual_by_true_output.get(c,0)
            if len(self.projected_input) == 0:
                proportion_effectif = 0
            else:
                proportion_effectif = effectif / float(len(self.projected_input)) * 100

            values.append(['{0:.2f}'.format(accuracy)+"%", '{0:.0f} ({1:.2f}%)'.format(effectif, proportion_effectif)])

        summary = ax.table(
            cellText=values,
            rowLabels=[' '+str(r[:6])+' ' for r in row_labels],
            colLabels=cols,
            loc='center',
        )
        summary.auto_set_font_size(False)
        summary.set_fontsize(8)
        logging.info("global_summary=ready")
    
    def request_heatmap(self, method, axe, title):
        """
        Draw a heatmap based on a heatmap_builder on an axe

        :param heatmap_builder: a Vizualization parameterless method which returns patches
        :param axe: matplotlib axe object in which the heatmap will be plotted
        """
        heatmap_constructor = heatmap.make_heatmap(
                method=method,
                )
        this_heatmap = heatmap_constructor(
                vizualization=self,
                )

        self.heatmaps.append((this_heatmap, axe, title))

    def export(self, output_path, format='csv'):
        """
        Export your selected data in a .csv file for analysis
        """
        logging.info('exporting:...')
        if output_path == "":
            logging.warn("No export name specified, aborting")
            return

        columns = [
                'prediction',
                'projected coordinates',
                'is_correct',
                ]

        rows =    [
                [
                    self.projected_input[idx],
                    self.prediction_outputs[idx],
                    int(idx in self.index_good_predicted),
                    ]
                for idx,c in enumerate(self.cluster_by_idx)
                if c in self.selected_clusters
                ]

        if self.x_raw.any():
            columns = [*self.x_raw_columns, *columns]
            rows =  [
                        [
                            *self.x_raw[idx],
                            self.prediction_outputs[idx],
                            self.projected_input[idx],
                            int(idx in self.index_good_predicted),
                            ]
                        for idx,c in enumerate(self.cluster_by_idx)
                        if c in self.selected_clusters
                        ]

        to_export =  [columns, *rows]

        if format=='csv':
            wr = csv.writer(open(output_path, 'w'))
            wr.writerows(to_export)
            logging.info('exporting: done')
    
        else:
            logging.warn('exporting: format not available')

    def make_clusterization_name(self, name_clusters):
        """
        Filename containing the saved session
        """
        return name_clusters + '#' + self.version + '.pkl'

    def save_clusterization(self, name_clusters='clusters'):
        """
        Basically saves a clusterizer and replays a sequence of leftclicks
        """
        if (' ' in name_clusters) or ('.' in name_clusters):
            self.viz_handler.pop_dialog("Please do not use space or '.' in name!")
            return

        filename  = self.make_clusterization_name(name_clusters)
        session   = (self.version, self.clusterizer, self.left_clicks) 
        full_path = os.path.join(self.saved_clusters_path, filename)

        logging.info("user clusters: saving in {}".format(full_path))
        pickle.dump(
                session,
                open(full_path, 'wb')
                )
        self.viz_handler.user_cluster_menulist.add_items([filename])
        return filename

    def load_clusterization(self, name_clusters):
        """
        Basically loads clustering engine and sequence of left_clicks
        """
        full_path = os.path.join(self.saved_clusters_path, name_clusters)

        version_to_load, clusterizer, left_clicks_to_reproduce = pickle.load(
                open(full_path, 'rb'),
                )
        
        if version_to_load != self.version:
            self.viz_handler.pop_dialog(
                    "Impossible to load clusters : bad VERSION (dataset do not match)")
            return
        
        self.reset_viz()
        self.left_clicks = set()
        self.clusterizer = clusterizer
        self.init_clusters()
        self.reset_summary()

        for left_click in left_clicks_to_reproduce:
            self.do_left_click(left_click)

        self.update_summary()
        self.print_summary(self.summary_axe)
        self.print_global_summary(self.global_summary_axe)
        
        self.update_all_heatmaps()
        self.request_new_frontiers('none')
        self.refresh_graph()

    
    def clear_cluster_view(self):
        for clr_view in self.cluster_view:
            clr_view.clear()

    def add_cluster_view(self, feature_to_display, plotter_names):
        logging.info(
                "cluster_view=adding a feature to display, {}:{}".format(
                    feature_to_display, plotter_names)
                )
        self.cluster_view.append(
                Cluster_viewer(
                    {feature_to_display:plotter_names},
                    self.x_raw,
                    self.x_raw_columns,
                    show_dichotomy=True,
                    )
                )
        self.additional_figures.append(self.cluster_view[-1])
        self.viz_handler.add_figure(
                self.additional_figures[-1],
                "Cluster view: {}".format(plotter_names[0])
                )
        logging.info("cluster_view=ready")
        if self.selected_clusters:
            self.update_cluster_view()

    def update_cluster_view(self):
        if self.cluster_view:
            logging.info('cluster_viewer=sorting indexes')
            index_selected  = [
                    idx for clr in self.selected_clusters for idx in self.index_by_cluster_label.get(clr,[])]
            index_good = [idx for idx in index_selected if idx in self.index_good_predicted]
            index_bad  = [idx for idx in index_selected if idx in self.index_bad_predicted ]

            for clr_view in self.cluster_view:
                logging.info('cluster_viewer=plotting sorted indexes')
                clr_view.update_cluster_view(
                        index_good    = index_good,
                        index_bad     = index_bad,
                        data_by_index = self.x_raw,
                        )
            logging.info('cluster_viewer=ready')

    def plot(self):
        """
        Plot the Vizualization, define axes, add scatterplot, buttons, etc..
        """
        self.init_clusters()
        self.main_fig = matplotlib.figure.Figure()
        #self.cluster_view = matplotlib.figure.Figure()
        self.cluster_view       = []
        self.additional_figures = []

        gs=GridSpec(3,4)
        #self.view_details = View_details(self.x_raw)
        self.viz_handler = Viz_handler(
                viz_engine         = self,
                figure             = self.main_fig,
                onclick            = self.onclick,
                additional_figures = self.additional_figures,
                )

        for feature_to_display,plotter_name in self.features_to_display.items():
            self.add_cluster_view(feature_to_display, plotter_name)

        # main subplot with the scatter plot
        self.ax = self.main_fig.add_subplot(gs[:2,:3])
        self.ax_base_title = 'Correct VS incorrect predictions'
        self.ax.set_title(self.ax_base_title)

        # summary_subplot with table of local stats
        self.summary_axe = self.main_fig.add_subplot(gs[2,:3])
        self.summary_axe.axis('off')

        self.global_summary_axe = self.main_fig.add_subplot(gs[2,3])
        self.global_summary_axe.axis('off')
        self.print_global_summary(self.global_summary_axe)

        # heatmap subplots
        # contain accuracy of correct prediction and entropy

        self.axes_needing_borders = [self.ax]

        # draw heatmap
        logging.info("heatmap=calculating")
        # self.cluster_view = self.main_fig.add_subplot(gs[1,3])
        

        self.heatmaps = []
        
        self.heatmap0 = self.main_fig.add_subplot(
                gs[1,3],
                sharex=self.ax,
                sharey=self.ax)
        self.request_heatmap(
                method=self.heatmaps_requested[0],
                axe=self.heatmap0,
                title='Heatmap: accuracy correct predictions')
        self.axes_needing_borders.append(self.heatmap0)
        

        self.heatmap1 = self.main_fig.add_subplot(
                gs[0,3],
                sharex=self.ax,
                sharey=self.ax)
        self.request_heatmap(
                method=self.heatmaps_requested[1],
                axe=self.heatmap1,
                title='Heatmap: cross-entropy Cluster-All')
        self.axes_needing_borders.append(self.heatmap1)
        
        self.update_all_heatmaps()
        logging.info("heatmap=ready")

        # draw scatter plot
        self.reset_viz()

        self.request_new_frontiers('none')
        logging.info('Vizualization=ready')
        self.viz_handler.is_ready()

    def show(self):
        logging.info('showing main window')
        self.viz_handler.show()
        logging.info('showing cluster diving window')
        #self.cluster_diver.show()
        logging.info("showing done")
        sys.exit(self.viz_handler.app.exec_())

    def refresh_graph(self):
        logging.info('refreshing main window')
        self.viz_handler.refresh()
        logging.info("refreshing done")



###########################################################################################
#
# How to add a new frontier engine (aka similarity measure) ?
#
##################################
#
# Simply creates a class in a new module that
# implements vizuka.similarity.similarity.Similarity
#
# Below is a simple example, to draw all frontiers (alpha is 1)
#
########################################################################



#######################################################
# WORKING EXAMPLE
##################
#
# from vizuka.similarity import similarity
#
# class AllFrontiers(similarity.Similarity):
#     def __init__(self):
#         pass
#     def compare(dict0, dict1, force_inside01=True):
#         """
#         Draw all frontiers
#         """
#         return 1 # Here 1 means "solid", 0 is "invisible. .5 for mid-solid



"""
This is the main script to launch everything
Is able to reduce the data and launch a vizualization
from it.
"""

import logging
import os
import argparse

import matplotlib
matplotlib.use('Qt5Agg')  # noqa
import numpy as np
from pyfiglet import Figlet

from vizuka import data_loader
from vizuka import vizualization
from vizuka import launch_reduce
from vizuka import cluster_viewer
from vizuka import heatmap

logger = logging.getLogger()
logger.setLevel(logging.WARN)

def main():

    """
    See --help if you want help
    """
    
    from vizuka.config import (
        BASE_PATH,
        DATA_PATH,
        VERSION,
        REDUCED_DATA_PATH,
        MODEL_PATH,
        INPUT_FILE_BASE_NAME,
        PROJECTION_DEFAULT_PARAMS,
        DEFAULT_PROJECTOR,
        path_builder,
    )

    print(Figlet().renderText('Vizuka'))

    plotters_builtin, plotters_extra = cluster_viewer.list_plotter()
    plotters_available = {**plotters_builtin, **plotters_extra}
    
    heatmaps_builtin, heatmaps_extra = heatmap.list_heatmap()
    heatmaps_available = {**heatmaps_builtin, **heatmaps_extra}

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--mnist', action='store_true',
         help='download, fit, and vizualize MNIST dataset.')
    parser.add_argument(
            '--show-required-files', action='store_true',
            help='show all the required files for optimal use',
            )
    parser.add_argument(
        '-p', '--path',
         help='change the location of your data/ folder, and auto-creates data/set/ data/reduced/ data/graph/ data/models/')
    parser.add_argument(
        '-s', '--feature_to_show', action='append',
        help='(available later in IHM) Usage: -s MY_COLUMN_NAME:PLOTTER with COLUMN_NAME with value in'
             ' data/set/raw_data.npz["columns"] (see --show-required-files)'
             ' and PLOTTER being a value in {}. Adds this non-preprocessed/human-readable feature to the cluster view'.format(list(plotters_available.keys())))
    parser.add_argument(
        '-h1', '--heatmap1',
        help='Specify the 1st heatmap to show, heatmaps available:{}'.format(list(heatmaps_available.keys())))
    parser.add_argument(
        '-h2', '--heatmap2',
         help='Specify the 2nd heatmap to show')
    parser.add_argument(
        '-v', '--version', type=str,
        help='(optional) specify a version of the files to load/generate (see vizuka --show-required-files), currently: '+VERSION)
    parser.add_argument(
        '--force-no-predict', action="store_true",
        help='(not recommended) do not load a predictions file : vizualize as if you predicted with 100percent accuracy')
    parser.add_argument(
        '--no-vizualize', action="store_true",
         help='(for debug) do not prepare a nice data vizualization')
    parser.add_argument(
        '--no-plot', action="store_true",
         help='(for debug) do not show a nice data vizualization (but prepare it nonetheless)')
    parser.add_argument(
            '--verbose', action="store_true",
            help="verbose mode")
    
    parser.set_defaults(
            heatmap1 ='accuracy',
            heatmap2 ='entropy',
            no_plot             =False,
            no_vizualize        =False,
            show_required_files =False,
            force_no_predict    =False,
            mnist               =False,
            version =VERSION,
            path    =os.path.join(os.path.dirname(__file__),BASE_PATH),
            feature_to_show   =[],
            verbose = False,
            )

    args = parser.parse_args()
    path                 = args.path

    (
        DATA_PATH,
        REDUCED_DATA_PATH,
        MODEL_PATH,
        _,
        _,
        _,
        ) = path_builder(path)

    no_vizualize = args.no_vizualize
    no_plot      = args.no_plot
    version      = args.version
    features_name_to_display = args.feature_to_show
    force_no_predict = args.force_no_predict
    show_required_files=args.show_required_files
    heatmap1    = args.heatmap1
    heatmap2    = args.heatmap2
    verbose     = args.verbose

    print(
            'Generate 2D projections with cmd "vizuka-reduce"\n'
            "VERSION: Loading dataset labeled {} (cf --version)".format(version)
            )

    if verbose:
        logger.setLevel(logging.DEBUG)

    if args.show_required_files:
        print(
        '\n\nVizuka needs the following files :\n\n'
        '\nVERSION: string that identifies your dataset (default is MNIST_example)\n'
        'PATH\t: data/ is located in '+str(os.path.dirname(__file__))+', change with --path\n'
        'FORMAT\t: all are .npz file\n\n'
        'REQUIRED:\n'
        '=========\n'
        '\t + data/set/preprocessed_inputs_VERSION.npz\n'
        '\t ------------------------------------------\n'
        '\t\t x:\t preprocessed inputs, your feature space\n'
        '\t\t y:\t outputs to be predicted, the "true" class\n'
        '\t\t NB:\t this is the only mandatory file, the following is highly recommended:\n'
        '\n\n'
        'OPTIONAL BUT USEFUL:\n'
        '===================\n'
        '\t + data/models/predict_VERSION.npz -> optional but recommended\n'
        '\t -------------------------------------------------------------\n'
        '\t\t y:\t predictions returned by your algorithm\n'
        '\t\t NB:\t should be same formatting as in preprocessed_inputs_VERSION["y"])\n'
        "\t\t\t if you don't have one, use --force-no-predict\n"
        '\n\n'
        '\t + data/set/raw_data.npz -> optional\n'
        '\t --------------------------\n'
        '\t\t x:\t\t array of inputs BEFORE preprocessing\n'
        '\t\t columns:\t the name of the columns variable in x\n'
        '\t\t NB:\t this file is used if you run vizuka with\n'
        '\t\t\t    --feature-name-to-display COLUMN_NAME:PLOTTER or\n'
        '\t\t\t    in IHM -> Cluster exploration / Add cluster info\n'
        '\t\t\t    (see help for details)\n'
        '\n\n'
        'GENERATED BY VIZUKA:\n'
        '====================\n'
        '\t + data/reduced/algoname#VERSION#PARAM1_NAME::VALUE#PARAM2_NAME::VALUE.npz\n'
        '\t ------------------------------------------------------------------------\n'
        '\t\t x2D:\t projections of the preprocessed inputs x in a 2D space\n'
        '\t\t NB:\t you can change default projection parameters and works with several ones\n'      '\t\t\t see vizuka-reduce'
        '\n\nSee '+str(os.path.join(os.path.dirname(__file__), 'data'))+' for real example, if you doubt\n'
        )
        return

    
    if args.mnist:
        from vizuka.example import load_mnist
        version = load_mnist.version
        features_name_to_display = ['image:images']

    
    new_fntd = {}
    error_msg  = 'Argument should be COLUMN_NAME:PLOTTER with plotter in {}'.format(list(plotters_available.keys()))
    e = Exception(error_msg)
    for feature_name in features_name_to_display:
        if ':' not in feature_name:
            raise e
        k,v = feature_name.split(':')
        # TODO if v not in cluster_diving.plotter.keys():
        #    raise e
        plotters = new_fntd.get(k, [])
        plotters.append(v)
        new_fntd[k] = plotters
    features_name_to_display = new_fntd

    logger.info("Starting script")
    logger.info("preprocessed_dataset=loading")

    (
        x,
        y,
        class_encoder,
        class_decoder,
        loaded,
        preprocessed_filename,
    ) = data_loader.load_preprocessed(
            file_base_name   = INPUT_FILE_BASE_NAME,
            path             = DATA_PATH,
            version          = version,
            )

    if not loaded:
        logging.warn("No data found\nCorresponding file not found: {}\nPlease check --show-required-files".format(preprocessed_filename))
        if version=='MNIST_example':
            print(
                    '\n\nIf you are trying to load the MNIST example, you need to launch'
                    ' "vizuka --mnist" in order to download the test data (code in vizuka/example/)'
                    )
        return

    logger.info('preprocessed_dataset=loaded')

    projections_available = data_loader.list_projections(REDUCED_DATA_PATH, version=version)

    if len(projections_available)==0:
        logger.warn("2D_dataset=No reduced data found! Please use vizuka-reduce to generate some")
        return

    choice_list, choice_dict = "", {}

    for i,(method, version_, params), in enumerate(projections_available):
        param_str = ''.join(["\t\t\t{}: {}\n".format(name, value) for name, value in params.items()])
        choice_list+="\t [{}]: \t{}\n\t\tparameters:\n{}\n".format(i, method, param_str)
        choice_dict[i]=method
    choice = input( "Projections datasets available: (generate more with vizuka-reduce)"
            ")\n"+choice_list+"Which dataset number to load?\n\n\t[?] > ")
    try:
        choice_int=int(choice)
    except ValueError:
        logging.warn("Please enter a valid dataset number! (e.g: 0)\nABORTING")
        return
    
    try:
        selected_method      = choice_dict[choice_int]
        _, _, selected_params = projections_available[choice_int]
    except KeyError:
        logging.warn("Don't be silly, enter a valid int number (between 0 and {})".format(len(choice_dict)-1))
        logging.warn("ABORTING")
        return

    x_2D = data_loader.load_projection(
        algorithm_name        =   selected_method,
        parameters            =   selected_params,
        version               =           version,
        path                  = REDUCED_DATA_PATH,
    )
    
    ###############
    # PREDICT
    if force_no_predict:
        x_predicted = y
    else:
        try:
            logger.info('predictions=loading')
            x_predicted = data_loader.load_predict(
                path=MODEL_PATH,
                version=version,
            )
            logger.info('predictions=ready')
        except FileNotFoundError:
            logger.warn((
                    "Nothing found in {}, no predictions to vizualize\n"
                    "if this is intended you can force the vizualization"
                    "with --force_no_predict :\n{}\n"
                    ).format(MODEL_PATH, os.listdir(MODEL_PATH)))
            return
    
    raw = data_loader.load_raw(version, DATA_PATH)

    if raw:
        logger.info("loading raw transactions for analysis..")
        raw_data, raw_columns = raw
    else:
        logger.info('no raw data provided, all cluster vizualization disabled! (-s and -f options)')
        raw_data    = np.array([])
        raw_columns = []
        features_name_to_display = {}


    
    if not no_vizualize:

        f = vizualization.Vizualization(
            projected_input=x_2D,
            predicted_outputs=x_predicted,
            raw_inputs=raw_data,
            raw_inputs_columns=raw_columns,
            correct_outputs=y,
            resolution=200,
            class_decoder=class_decoder,
            class_encoder=class_encoder,
            special_class='x',
            nb_of_clusters=120,
            features_name_to_display = features_name_to_display,
            heatmaps_requested = [heatmap1, heatmap2],
            output_path=os.path.join('output.csv'),
            base_path=path,
            version=version,
        )

        if not no_plot:
            f.plot()
            f.show()
    return f

if __name__ == '__main__':
    main()



"""
Qt handler for drawing buttons and IHM black magic.
Please do read it you fool. I must admit I am not
proud of everything written down there.

This should be rewritten with QtCreator's help.
"""
import sys
import logging
import os

import matplotlib
matplotlib.use('Qt5Agg')  # noqa
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
import matplotlib.gridspec as gridspec
from matplotlib import pyplot as plt
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtWidgets import (
    QApplication,
    QMainWindow,
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QSizePolicy,
    QDockWidget,
    QLineEdit,
    QInputDialog,
    QMessageBox,
    QAction,
    QActionGroup,
    QMenu,
    QFileDialog,
    QDialog,
    QRadioButton,
    QLabel,
)
import numpy as np

from vizuka.graphics import qt_helpers
from vizuka import clustering
from vizuka import frontier
from vizuka.cluster_viewer import make_plotter, list_plotter


def onclick_wrapper(onclick):
    """
    This decorator for onclick detects if the mouse event
    should trigger something.
    I tried to make it worst but this is my best
    """
    def wrapper(*args, **kwargs):
        if args[0].detect_mouse_event:  # args[0] is self
            return onclick(*args, **kwargs)
        else:
            return lambda x: None
    return wrapper

CLUSTER_PLOTTER = {}
            
class Cluster_viewer(matplotlib.figure.Figure):

    def __init__(self, features_to_display, x_raw, x_raw_columns, show_dichotomy=True):
        super().__init__()
        self.x_raw = x_raw
        self.x_raw_columns = x_raw_columns
        self.show_dichotomy = show_dichotomy # deprecated

        self.features_to_display = features_to_display
        self.spec_by_name = {}
        self.cluster_view_selected_indexes = []

        self.spec = gridspec.GridSpec(
                len(features_to_display.keys()),
                2, wspace=0.2
                )

        for idx,feature_name in enumerate(features_to_display.keys()):
            for plotter in features_to_display[feature_name]:
                self.spec_by_name[feature_name+plotter] = {}
                self.spec_by_name[feature_name+plotter]['good'] = self.spec[idx%2]
                self.spec_by_name[feature_name+plotter]['bad' ] = self.spec[idx%2+1]

                if plotter not in CLUSTER_PLOTTER.keys():
                    CLUSTER_PLOTTER[plotter] = make_plotter(plotter)
        
        def plot_help(plotter, plotter_name, data_name, fig, spec_to_update_, key, y):

            spec_to_update = spec_to_update_[key]
            axe = plt.Subplot(fig, spec_to_update)
            axe.axis("off")
            text = (
                    'Here will be displayed the distribution of the column {}\n'.format(data_name)
                    +"The visualisation tool '{}' will be used\n\n".format(plotter_name)
                    +'Click on a cluster to begin'
                    )
            axe.text(
                    0.5, 0.5,
                    text,
                    horizontalalignment='center',
                    verticalalignment=y,
                    )
            fig.add_subplot(axe)

            if 'log' in plotter_name:
                data_name += ' - log'
            data_name +=  ' - {} predictions'.format(key)
            if axe:
                axe.set_title(data_name)

        for key in ['good', 'bad']:
            for data_name in self.features_to_display:
                for plotter_name in self.features_to_display[data_name]:
                    plotter = CLUSTER_PLOTTER[plotter_name]
                    spec_to_update = self.spec_by_name[data_name+plotter_name]
                    y = 'top' if key=='good' else 'bottom'
                    plot_help(plotter, plotter_name, data_name, self, spec_to_update, key, y)

    def update_cluster_view(self, index_good, index_bad, data_by_index):
        """
        Updates the axes with the data of the selected cluster

        index_good: indexes of all good predictions
        index_bad: indexes of all bad predicitons
        data_by_index: data
        """
        self.clear()
        self.cluster_view_selected_indexes = set(index_good+index_bad)
        
        
        selected_xs_raw  ={}
        data_to_display = {}

        if self.show_dichotomy:
            selected_xs_raw['good'] = [
                    data_by_index[idx] for idx in self.cluster_view_selected_indexes if idx in index_good]
            selected_xs_raw['bad' ] = [
                    data_by_index[idx] for idx in self.cluster_view_selected_indexes if idx in index_bad ]
        else:
            selected_xs_raw['all'] = [
                    data_by_index[idx] for idx in self.cluster_view_selected_indexes]

        columns_to_display = [list(self.x_raw_columns).index(i) for i in self.features_to_display]
        if self.show_dichotomy:
            data_to_display['good'] = {
                self.x_raw_columns[i]:[x[i] for x in selected_xs_raw['good']]
                for i in columns_to_display
                }
            data_to_display['bad'] = {
                self.x_raw_columns[i]:[x[i] for x in selected_xs_raw['bad']]
                for i in columns_to_display
                }
        else:
            data_to_display['all'] = {
                    self.x_raw_columns[i]:[x[i] for x in selected_xs_raw['all']]
                    for i in columns_to_display
                    }

        def plot_it(plotter, plotter_name, data_to_display, data_name, fig, spec_to_update_, key):

            spec_to_update = spec_to_update_[key]
            data = data_to_display[key][data_name]
            axe = plotter(data, fig, spec_to_update)
            if 'log' in plotter_name:
                data_name += ' - log'
            data_name +=  ' - {} predictions'.format(key)
            if axe:
                axe.set_title(data_name)

        for key in ['good', 'bad']:
            for data_name in self.features_to_display:
                for plotter_name in self.features_to_display[data_name]:
                    plotter = CLUSTER_PLOTTER[plotter_name]
                    spec_to_update = self.spec_by_name[data_name+plotter_name]
                    plot_it(plotter, plotter_name, data_to_display, data_name, self, spec_to_update, key) 

        
    def reset(self):
        self.clear()
        self.cluster_view_selected_indexes = []


class Qt_matplotlib_handler():

    def __init__(self, figure):
        self.right_dock = QtCore.Qt.RightDockWidgetArea
        self.textboxs = {}
        self.plottings = []

        # configure the app + main window
        self.app = QApplication(sys.argv)
        self.window = QMainWindow(parent=None)
        self.features_to_filter=None
        self.figure = figure


    def show(self):
        """
        Shows the window you built with your tears
        """
        self.window.showMaximized()
        for win in self.additional_windows:
            win.show()
        self.refresh()
        logging.info("refreshed")
        logging.info("showed")

    def refresh(self):
        """
        Refresh the matplotlib plots
        """
        for plotting in self.plottings:
            try:
                plotting.canvas.draw()
            except ValueError as e:
                # This exception is raised when nothing to be draw
                logging.warn(e)
                for i in self.figure.get_axes():
                    i.clear()
                
    
    @onclick_wrapper
    def onclick(self, *args, **kwargs):
        """
        ..seealso::onclick_wrapper
        """
        self.base_onclick(*args, **kwargs)

        
    def toogle_detect_mouse_event(self, *args, **kwargs):
        """
        Enable/Disable mouse_event handling ..seealso::onclick_wrapper
        """
        self.detect_mouse_event = not self.detect_mouse_event


class Viz_handler(Qt_matplotlib_handler):
    """
    A Viz_handler is attached to the viz_engine defined in vizualization.py
    It basically lists lots of QWidget things and propose methods to init
    them "gracefully", hum...

    Only IHM here.
    """
    
    def __init__(
            self,
            viz_engine,
            figure,
            onclick,
            additional_figures=[],
            ):
        """
        This object is a QtWindow (or 2-3-4...) with a mamtplotlib.Figure
        and a onclick event handler. As you can see it is also linked to
        a viz_engine which will do the magic.
        
        This instance should only do IHM stuff, nothing intellectual, everything
        is handled by the viz_engine
        """
        super(Viz_handler, self).__init__(figure)

        self.viz_engine = viz_engine
        self.detect_mouse_event = True
        self.base_onclick = onclick
        self.window.setWindowTitle('Data vizualization')
        self.additional_figures = additional_figures
        self.all_custom_filters = {}

        self.additional_windows = []

        # add the main figure
        qt_helpers.add_figure(self.figure, window=self.window, plottings=self.plottings, onclick=self.onclick)
        #self.cluster_diver = Cluster_diver(self.x_raw, self.x_raw_columns, ['ape_code'])

        for fig in self.additional_figures:
            self.add_figure(fig, 'Cluster viewer')

        # self.add_figure(self.additional_figure, onclick=None, window=self.additional_window)
        self.initUI()

    def add_figure(self, fig, name):
        self.additional_windows.append(qt_helpers.add_window(self.window, name))
        qt_helpers.add_figure(fig, self.additional_windows[-1], plottings=self.plottings)
        self.additional_windows[-1].show()

    
    def initUI(self):
        window = self.window
        
        window.statusBar().showMessage('Loading...')

        menubar  = window.menuBar()

        qt_helpers.add_simple_menu(
                "Load",
                {"Load a set of predictions":self.load_prediction},
                menubar,
                window,
                )
        qt_helpers.add_simple_menu(
                'Export',
                { 'Export data in selected clusters in csv':self.export },
                menubar,
                window,
                )

        clusteringMenu = menubar.addMenu("Clustering")

        saveclustersAct = QAction('Save current clusters', window)
        saveclustersAct.triggered.connect(self.save_clusters)
        clusteringMenu.addAction(saveclustersAct)

        requestClusteringMenu = QMenu('Request clustering', window)
        builtin_cl, extra_cl = clustering.list_clusterizer()
        available_clustering_engines = {**builtin_cl, **extra_cl}
        for clustering_engine_name in available_clustering_engines.keys():
            clusteringAct = QAction(clustering_engine_name, window)
            clusteringAct.triggered.connect(
                    lambda x, name=clustering_engine_name:self.request_new_clustering(name))
            requestClusteringMenu.addAction(clusteringAct)
        clusteringMenu.addMenu(requestClusteringMenu)

        self.loadClusterMenu = QMenu('Load saved clusters', window)
        for cluster_name in self.viz_engine.saved_clusters:
            loadAct = QAction(cluster_name, window)
            loadAct.triggered.connect(
                    lambda x, name=cluster_name:self.viz_engine.load_clusterization(name))
            self.loadClusterMenu.addAction(loadAct)
        clusteringMenu.addMenu(self.loadClusterMenu)
        
        frontierMenu = menubar.addMenu('Cluster frontiers')
        builtin_fr, extra_fr = frontier.list_frontiers()
        available_frontiers = {**builtin_fr, **extra_fr}
        for frontier_name in available_frontiers.keys():
            frontierAct = QAction(frontier_name, window)
            frontierAct.triggered.connect(
                    lambda x, name=frontier_name:self.viz_engine.request_new_frontiers(name))
            frontierMenu.addAction(frontierAct)
        
        self.filterMenu = menubar.addMenu('Filters')

        trueClassFilter       = QMenu('Filter by true class', window)
        self.all_true_filters = []

        for cls in self.viz_engine.possible_outputs_list:
            name = cls if len(cls)<30 else str(cls[:29])+'...'
            f = QAction(name, window, checkable=True)
            f.setChecked(True)
            f.class_ = cls
            self.all_true_filters.append(f)
            f.triggered.connect(lambda x:self.viz_engine.filter_by_correct_class(self.all_true_filters))
            trueClassFilter.addAction(f)
        
        def select_all_true(boolean=True):
            for f in self.all_true_filters:
                f.setChecked(boolean)
            self.viz_engine.filter_by_correct_class(self.all_true_filters)

        select_all = QAction("Select all", window)
        select_all.triggered.connect(lambda x:select_all_true(True))
        trueClassFilter.addAction(select_all)
        
        select_none = QAction("Unselect all", window)
        select_none.triggered.connect(lambda x:select_all_true(False))
        trueClassFilter.addAction(select_none)

        self.filterMenu.addMenu(trueClassFilter)
        
        predictClassFilter       = QMenu('Filter by predicted class', window)
        self.all_predicted_filters = []

        for cls in self.viz_engine.possible_outputs_list:
            name = cls if len(cls)<30 else str(cls[:29])+'...'
            f = QAction(name, window, checkable=True)
            f.setChecked(True)
            f.class_ = cls
            self.all_predicted_filters.append(f)
            f.triggered.connect(lambda x:self.viz_engine.filter_by_predicted_class(self.all_predicted_filters))
            predictClassFilter.addAction(f)

        def select_all_predicted(boolean=True):
            for f in self.all_predicted_filters:
                f.setChecked(boolean)
            self.viz_engine.filter_by_predicted_class(self.all_predicted_filters)

        select_all_p = QAction("Select all", window)
        select_all_p.triggered.connect(lambda x:select_all_predicted(True))
        predictClassFilter.addAction(select_all_p)
        
        select_none_p = QAction("Unselect all", window)
        select_none_p.triggered.connect(lambda x:select_all_predicted(False))
        predictClassFilter.addAction(select_none_p)

        self.filterMenu.addMenu(predictClassFilter)

        moreFilterAction = QAction("Add a filter from raw data", self.window)
        moreFilterAction.triggered.connect(lambda x:self.add_raw_filter())
        self.filterMenu.addAction(moreFilterAction)

        
        navigationMenu = menubar.addMenu('Navigation')
        mouseEnabledAction = QAction('Disable mouse click', window, checkable=True)
        mouseEnabledAction.triggered.connect(self.toogle_detect_mouse_event)
        navigationMenu.addAction(mouseEnabledAction)

        clusterviewMenu   = menubar.addMenu("Cluster exploration")
        clustersummaryAct = QAction("Add custom cluster info", window)
        clustersummaryAct.triggered.connect(self.request_cluster_viewer)
        clusterviewMenu.addAction(clustersummaryAct)

                
        colorMenu = menubar.addMenu("Color mode")
        color_modes = {
                'Colorize correctness of predictions':"by_prediction",
                'Colorize by classes':"by_true_class",
                }
        def on_color_radio_button_toggled(s):
            radiobutton = s.sender()
            if radiobutton.isChecked():
                self.viz_engine.request_color_mode(color_modes[radiobutton.column_name])
        radio_color_group = QActionGroup(colorMenu, exclusive=True)
        for i,mode in enumerate(color_modes.keys()):
            radiobutton = QAction(mode, radio_color_group, checkable=True)
            radiobutton.column_name = mode
            radiobutton.toggled.connect(lambda x:on_color_radio_button_toggled(radiobutton))
            if 'Colorize correctness of predictions' == mode:
                radiobutton.setChecked(True)
            colorMenu.addAction(radiobutton)
        
    
    def add_raw_filter(self):
        self.kikoo = AskColumnWindow(self.window, self.viz_engine, next_window=self.addCustomFilter)
        self.kikoo.show()

    def addCustomFilter(self,_, column_name, viz_engine):

        customFilter            = QMenu('Filter by {}'.format(column_name), self.window)
        column_index            = viz_engine.x_raw_columns.tolist().index(column_name)
        classes                 = set(np.array(viz_engine.x_raw)[:,column_index])
        self.all_custom_filters[column_index] = []

        for cls in classes:
            name = str(cls)
            f = QAction(name, self.window, checkable=True)
            f.setChecked(True)
            f.class_ = cls
            self.all_custom_filters[column_index].append(f)
            f.triggered.connect(lambda x:self.viz_engine.filter_by_feature(
                column_index,
                self.all_custom_filters,
                ))
            customFilter.addAction(f)

        def select_all_custom(boolean=True):
            for f in self.all_custom_filters[column_index]:
                f.setChecked(boolean)
            self.viz_engine.filter_by_feature(
                column_index,
                self.all_custom_filters,
                )

        select_all_c = QAction("Select all", self.window)
        select_all_c.triggered.connect(lambda x:select_all_custom(True))
        customFilter.addAction(select_all_c)
        
        select_none_c = QAction("Unselect all", self.window)
        select_none_c.triggered.connect(lambda x:select_all_custom(False))
        customFilter.addAction(select_none_c)

        self.filterMenu.addMenu(customFilter)
        
    def request_cluster_viewer(self):
        self.kikoo = AskColumnWindow(self.window, self.viz_engine, next_window=AskPlotterWindow)
        self.kikoo.show()

    def is_ready(self):
        self.window.statusBar().showMessage('Ready')

    def load_prediction(self):
        prediction_filename = self.file_dialog(
                "Load prediction",
                self.viz_engine.model_path,
                "{} Npz file (*{}.npz)".format(self.viz_engine.version, self.viz_engine.version),
                )
        if prediction_filename:
            self.viz_engine.reload_predict(prediction_filename)
        

    def is_loading(self):
        self.window.statusBar().showMessage('Loading...')

    def pop_dialog(self, msg):
        popup = QMessageBox(self.window)
        popup.setText(msg)
        popup.exec_()

    def file_dialog(self, name, directory, filter):
        dialog = QFileDialog(self.window, name, directory, filter)
        if dialog.exec_() == QDialog.Accepted:
            return dialog.selectedFiles()[0]


    def save_clusters(self):
        text, validated = QInputDialog.getText(
                self.window,
                "YOLO", "Name for the clusters you selected ?"
                )
        if validated:
            if text == '':
                text = "clusters"
            cluster_name = self.viz_engine.save_clusterization(text)
            loadAct = QAction(cluster_name, self.window)
            loadAct.triggered.connect(lambda:self.viz_engine.load_clusterization(cluster_name))
            self.loadClusterMenu.addAction(loadAct)

    def export(self):
        text, validated = QInputDialog.getText(
                self.window,
                "YOLO", "Name for the exported csv ?"
                )
        if validated:
            if text == '':
                text = "export.csv"
        self.viz_engine.export(text)

    def request_new_clustering(self, clustering_engine_name):
        self.clustering_params={}
        for requested_param in clustering.get_required_arguments(clustering_engine_name):
            text, validated = QInputDialog.getText(
                    self.window,
                    "YOLO", requested_param+' ?'
                    )
            self.clustering_params[requested_param]=float(text)

        return self.viz_engine.request_new_clustering(clustering_engine_name, self.clustering_params)


class AskPlotterWindow(QDialog):
    def __init__(self, parent, column_name, viz_engine):
        super(AskPlotterWindow, self).__init__(parent)

        layout = QVBoxLayout()
        self.setLayout(layout)
        self.viz_engine = viz_engine
        self.column_name = column_name
        
        builtin_pl, extra_pl = list_plotter()
        available_pl = {**builtin_pl, **extra_pl}

        layout.addWidget(QLabel("The following visualization tools are available"),0)
        layout.addWidget(QLabel("Which one do you want to use for exploration ?"),1)

        for i,(plotter_name, plotter_class) in enumerate(available_pl.items()):
            radiobutton = QRadioButton(plotter_name +' - '+plotter_class.get_help())
            radiobutton.plotter_name = plotter_name
            radiobutton.toggled.connect(self.on_radio_button_toggled)
            layout.addWidget(radiobutton, i+2)

        self.show()

    def on_radio_button_toggled(self):
        radiobutton = self.sender()
        if radiobutton.isChecked():
            self.viz_engine.add_cluster_view(self.column_name, [radiobutton.plotter_name])
            self.destroy()


class AskColumnWindow(QDialog):
    def __init__(self, parent, viz_engine, next_window):
        super(AskColumnWindow, self).__init__(parent)
        self.parent = parent
        
        self.next_window = next_window
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.viz_engine = viz_engine

        layout.addWidget(QLabel("Your raw_data_"+self.viz_engine.version+".npz file has the following column\n"),0)
        layout.addWidget(QLabel("Which one do you want to explore ?"),1)

        for i,column in enumerate(viz_engine.x_raw_columns):
            radiobutton = QRadioButton(column)
            radiobutton.column_name = column
            radiobutton.toggled.connect(self.on_radio_button_toggled)
            layout.addWidget(radiobutton, i+2)

    def on_radio_button_toggled(self):
        radiobutton = self.sender()
        if radiobutton.isChecked():
            self.kikoo = self.next_window(self.parent, radiobutton.column_name, self.viz_engine)
            self.destroy()



import logging
import os
import itertools

import numpy as np

from vizuka import dimension_reduction
from vizuka.dimension_reduction.projector import Projector
from vizuka.config import (
        MODEL_PATH,
        VERSION,
        DEFAULT_PREDICTOR,
        PROJECTION_DEFAULT_PARAMS,
        REDUCED_DATA_PATH,
        INPUT_FILE_BASE_NAME,
        DATA_PATH,
        RAW_NAME,
        )


def load_predict(path=MODEL_PATH, version=VERSION, namePredictor=DEFAULT_PREDICTOR):
    """
    Simply load the predictions associated with the VERSION data
    """
    logging.info("trying to load {}".format(path + namePredictor +'_'+ version + '.npz'))
    return np.load(path + namePredictor + version + '.npz')['y']

def load_predict_byname(filename, path=MODEL_PATH):
    """
    Simply load the predictions associated with the VERSION data
    """
    full_path = os.path.join(path, filename)
    logging.info("trying to load {}".format(full_path))
    return np.load(os.path.join(path, filename))['y']

def list_projections(reduced_path, version):
    """
    Returns a list containing [(algo_name, parameters),..] found in :param reduced_path:
    """
    files = [filename for filename in os.listdir(reduced_path) if (".npz" in filename) and (version in filename)]
    return  [Projector.get_param_from_name(filename[:-4]) for filename in files]

def load_projection(algorithm_name, parameters, version, path):
    logging.info("data_loader=loading projection:\n\talgorithm:{}\n\tparameters:{}".format(
                        algorithm_name,
                        parameters,
                        ))
    algo_builder = dimension_reduction.make_projector(algorithm_name)
    algo = algo_builder(**parameters)
    projection = algo.load_projection(version=version, path=path)

    logging.info("data_loader=ready")
    return projection

def load_raw(version, path):
    raw_filename = os.path.join(path, RAW_NAME + version + '.npz')
    if os.path.exists(raw_filename):
        raw_ = np.load(raw_filename)
        return raw_["x"], raw_["columns"]
    else:
        return None

def load_preprocessed(
        file_base_name=INPUT_FILE_BASE_NAME,
        path=DATA_PATH,
        version=VERSION,
):
    """
    Loads and returns the data for tSNE
    One-hotted and regular ML-encoding

    File to load should :
        - be in path
        - has name "(INPUT_FILE_BASE_NAME)_x_y_(VERSION).npz"
                (VERSION) is for e.g "_20170614"
                (FILE_BASE_NAME) if for e.g 'processed_predictions' or 'one_hot'
        - contains entry x with input data
        - contains entry y with output data (possible_outputs_list to predict)
        - optionnaly an encoder to translate machine-readable possible_outputs_list to human-readable possible_outputs_list
                (actually it is the opposite e.g: {604600:[False, True, False, .., False]})

    :return: (input for t-SNE, classes, encoder (humantomachine class possible_outputs_list),
    decoder (machinetohuman possible_outputs_list), loaded, path_of_filename)

    Note that encoder/decoder are assumed trivial if no encoder are found in npz

    """
    full_path = os.path.join(path, INPUT_FILE_BASE_NAME + version + '.npz')
    if not os.path.exists(full_path):
        loaded = False
        return np.array([]), np.array([]), lambda x:x, lambda x:x, loaded, full_path
    else:
        xy = np.load(full_path)
        loaded = True

    x, y = xy['x'], xy['y']
    
    class_encoder = {y: y for y in set(y)}
    class_decoder = lambda x: x # noqa

    x = xy['x']
    if 'y' in xy.keys():
        y = xy['y']
        del xy
        return (np.array(x), np.array(y), class_encoder, class_decoder, loaded, full_path)
    elif 'y' + '_decoded' in xy.keys():
        y_decoded = xy['y' + '_decoded']
        del xy
        return (np.array(x), np.array(y_decoded), class_encoder, class_decoder, loaded, full_path)




'''
Clustering engine to use with Vizualization

If you want to implement one, dont forget to add it
on qt_handler, to be able to select on the IHM
Also, adds it in make_clusterizer obviously

3 methods are necessary to implement, cf Clusterizer():
    init    - set main params
    fit     - to prepare the algo for the data
    predict - to find out the cluster of a (x,y)
'''
import pickle
import os
import logging

from matplotlib import pyplot as plt

from vizuka.config import (
        CACHE_PATH,
        NAME_VALUE_SEPARATOR,
        PARAMETERS_SEPARATOR,
        )

class Clusterizer():

        def get_param_from_name(name):
            """
            Returns the method and parameters of the clusterizer
            from its name
            """
            params = {}
            blocks = name.split(PARAMETERS_SEPARATOR)
            method = blocks[0]
            for block in blocks[1:]:
                if block:
                    name, value  = block.split(NAME_VALUE_SEPARATOR)
                    params[name] = value
            return method, params

        def __init__(self, **kwargs):
            """
            Builds a clusterizer object e.g: kmeans
            Do not know the datas at this point, just pass it the bare
            minimum to initialize an engine.

            :param *args **kwargs: parameters passed to the clusterizer engine
                                   it can literally be whatever you want
            """
            pass

        def register_parameters(self, parameters):
            """
            Register the parameters of the clusterizer to generate a name
            The name will be used for saving clusterizer in cache file (pickle)
            and replay session with Load clusterization in Vizuka main window

            :param parameters: a dict with the parameters of the clustering engine
            """
            self.parameters = parameters

        def get_name(self):
            """
            Returns a unique name for the clustering engine
            This name alone is not sufficient to find which cached clustering engine to
            reload as different datasets have a different clustering
            """
            ordered_parameters_name = list(self.parameters)
            ordered_parameters_name.sort()

            return ''.join([
                    self.method,PARAMETERS_SEPARATOR,
                    *[ str(name)+NAME_VALUE_SEPARATOR+str(self.parameters[name])+PARAMETERS_SEPARATOR for name in ordered_parameters_name ]
                    ])[:-1]

        def fit(self, xs):
            """
            First time the engine sees the data.
            Depending on the algorithm you may do your magic your own way
            and maybe store new variable in self, maybe store all the
            predicts for each x directly in a dict.

            :param xs: a list containing data to clusterize
            """
            pass


        def predict(self, xs):
            """
            Finds the cluster(s) in which the data is.

            :params xs: array-like of (x,y) points
            :return: array-like of cluster id
            """
            return (0, ) * len(xs)
        
        def loadable(self, version, path=CACHE_PATH):
            filename = ''.join(
                    [self.get_name(), PARAMETERS_SEPARATOR, version,'.pkl'])
            full_path = os.path.join(path, filename)

            if os.path.exists(full_path):
                logging.info('cluster: found clusters in {}'.format(full_path))
                return True
            else:
                logging.info('cluster: no cached clusters found in {}'.format(full_path))
                return False

        def save_cluster(self, version, path=CACHE_PATH):
            filename = ''.join(
                    [self.get_name(), PARAMETERS_SEPARATOR, version,'.pkl'])
            full_path = os.path.join(path, filename)
            logging.info('cluster: saving clusters in {}'.format(full_path))
            with open(full_path, 'wb') as f:
                pickle.dump(self, f)
        

        def load_cluster(self, version, path=CACHE_PATH):
            filename = ''.join(
                    [self.get_name(), PARAMETERS_SEPARATOR, version,'.pkl'])
            full_path = os.path.join(path, filename)
            logging.info('cluster: loading clusters in {}'.format(full_path))
            with open(full_path, 'rb') as f:
                loaded = pickle.load(f)
                self = loaded
                return loaded





if __name__ == '__main__':
    """
    Yes, I like to test my code with __main__
    """
    from vizuka import dim_reduction as dr
    from vizuka import data_loader as dl
    datas_sets, models = dl.load_tSNE()
    datas = datas_sets[50, 1000, 'pca', 15000]

    clusterizer = make_clusterizer(datas, method='kmeans', n_clusters=80)
    f, ax = plot_clusters(datas, clusterizer)

    plt.show()



from sklearn.cluster import KMeans

from vizuka.clustering.clusterizer import Clusterizer


class KmeansClusterizer(Clusterizer):

    def __init__(self, number_of_clusters=15):
        """
        Uses sklearn kmeans
        """
        self.register_parameters(
                parameters={'number_of_clusters':number_of_clusters})

        self.method = 'kmeans'
        self.number_of_clusters = number_of_clusters

        self.engine = KMeans(
                n_clusters=int(float(self.number_of_clusters)),
                )

    def fit(self, xs):
        """
        Fit the datas and find clusterization adapted to the data provided

        :param xs: data to clusterize
        """
        self.engine.fit(xs)

    def predict(self, xs):
        """
        Predicts cluster label

        :param xs: array-like of datas
        :return:   list of cluster possible_outputs_list
        """
        return self.engine.predict(xs)



from setuptools import setup, find_packages
import os

data = {
    'vizuka': [
        'data/set/.gitkeep',
        'data/models/.gitkeep',
        'data/cache/.gitkeep',
        'data/saved_clusters/.gitkeep',
        'data/reduced/.gitkeep',
        'example/tsne#MNIST_example#learning_rate::1000#n_iter::12000#perplexity::50.npz',
    ],
}

setup(
    name='Vizuka',
    version='0.36.1',
    packages=find_packages(),  #['vizuka/'],
    package_data=data,
    entry_points={
        'console_scripts': [
            'vizuka=vizuka.launch_viz:main',
            'vizuka-reduce=vizuka.launch_reduce:main',
        ],
    },
    description=
    'Represents your high-dimensional datas in a 2D space and play with it',
    long_description=open(
        os.path.join(os.path.dirname(__file__), 'README.md')).read(),
    install_requires=open(
        os.path.join(
            os.path.dirname(__file__),
            'vizuka/requirements/requirements.txt')).read(),
    license='GPL V3',
    author='Sofian Medbouhi',
    author_email='sof.m.sk@free.fr',
)



#############################################
#
#  How to add a dimension reduction algorithm ?
#
##################
#
# Below is a working example. You should simply inherits
# vizuka.dimension_reduction.projector.Projector and implements
# a contructor and a project method.
#
# Typically in the __init__ you instanciate your engine (e.g the
# TSNE object from sklearn.manifold)  and /!\ register its parameters
# (see example below). This is important for vizuka to be able to
# ask the users the values of the parameters and not confuse different
# projections with different parameters.
#
# "project" method takes as input an array of your high-dimensional
# and returns an array with projection (/!\: set the self.projections
# object to your 2D embedding array (see below)).
#
######################################################################

#####################################################################
#
# from vizuka.dimension_reduction import projector
# from sklearn.manifold import TSNE
# 
# class tSNE(projector.Projector):
# 
#     def __init__(self, perplexity, learning_rate, n_iter):
#         
#         self.method         = 'tsne'
# 
#         self.register_parameters(
#                 parameters = {
#                     'perplexity'   : perplexity,
#                     'learning_rate': learning_rate,
#                     'n_iter'       : n_iter,
#                     }
#                 )
# 
#         self.engine = TSNE(
#             perplexity    = self.parameters['perplexity'],
#             learning_rate = self.parameters['learning_rate'],
#             n_iter        = self.parameters['n_iter'],
#             **multicore_parameters,
#             )
#         self.projections = []
# 
#     def project(self, x):
#         self.projections = self.engine.fit_transform(x)
#         return self.projections



"""
Here are the default parameters used in all the package
There are 1.path 2.filenames 3.learning parameters
"""
import os

#
# ALL DEFAULT PATH
#


def path_builder(base_path):
    folders = [
        os.path.join(base_path, relative) for relative in [
            'set/',
            'reduced/',
            'models/',
            'graph/',
            'cache/',
            'saved_clusters/',
        ]
    ]
    for folder in folders:
        if not os.path.exists(folder):
            os.makedirs(folder)
    return folders


# Build all path from one base
BASE_PATH = os.path.join(os.path.dirname(__file__), 'data/')

(
    DATA_PATH,
    REDUCED_DATA_PATH,
    MODEL_PATH,
    GRAPH_PATH,
    CACHE_PATH,
    SAVED_CLUSTERS_PATH,
) = path_builder(BASE_PATH)

#
# ALL DEFAULT FILENAME
#

# File containing data to be t-SNEed
INPUT_FILE_BASE_NAME = 'preprocessed_inputs_'
RAW_NAME = 'raw_data_'

# default RN for predictions
DEFAULT_PREDICTOR = 'predict_'

# PROJECTIONS DATA will have a named generated from the algo projector parameters

# A version is a string added to the end of each filename
VERSION = 'MNIST_example'

#
#  LEARNING PARAMETERS
#

# Dimension reduction default parameters :
DEFAULT_PROJECTOR = 'tsne'
PROJECTION_DEFAULT_PARAMS = {
    'tsne': {
        'perplexity': 50,
        'learning_rate': 1000,
        'n_iter': 12000,
    },
    'pca': {
        'nb_dimension': 2,
        'min_ratio_variance_explained': -1,
    },
}

NAME_VALUE_SEPARATOR = '@@'
PARAMETERS_SEPARATOR = '#'
